#!/usr/bin/env python
# A script to record two video files concurrently:
#   - A screencast of the user's desktop, accompanied by audio from the microphone
#   - Video streamed from the webcam.

import os
import os.path
dbus_enabled = False
try:
	import dbus # For screensaver inhibition
	dbus_enabled = True
except ImportError:
	pass
import re
import sys
import subprocess
import signal

def signal_handler(signum, frame):
	print "Caught SIGINT, shutting down"

def prompt(choices, prompt_string=None, default=None):
	"""Accepts a list of choice strings, with an optional prompt string and (zero-indexed) default choice.  Prompts the user and returns the (zero-indexed) index of the users (one-indexed) choice"""
	answer_satisfactory = False
	n = len(choices)
	while True:
		if not prompt_string:
			prompt_string = "Please select one of the following %d choices:" % n
		print prompt_string
		print
		for i in xrange(n):
			print "\t[%d] - %s" % (i+1, choices[i])
		print
		if default:
			# Verify that default is in the acceptable range
			if int(default) > n or int(default) < 1:
				default = None
		reqstring = "".join(["Your choice ", "[%s]" % str(default+1) if default else "" , ": "])
		print reqstring,
		answer = sys.stdin.readline().strip()
		if answer == "":
			if default:
				return int(default)
			else:
				continue
		else:
			try:
				index = int(answer)
				if index >= 1 and index <= n:
					return index-1
				else:
					print "%d is not a valid choice.  Please enter and integer between 1 and %d" % (index, n)
			except ValueError:
				print "I didn't understand that."

def choose_monitor():
	"""Returns a tuple (name, width, height, x offset, y offset) describing the desired screen area to record."""
	xrandrlines = subprocess.check_output("xrandr").split("\n")

	activedisplays = []
	import re
	for line in xrandrlines:
		match = re.search("(.+) connected (\d+)x(\d+)\+(\d+)\+(\d+)", line)
		if match:
			activedisplays.append( match.groups() )
	# activedisplays is now a list of tuples (title, w, h, x, y)
	strings = ["%s @ %sx%s+%s+%s" % tup for tup in activedisplays]

	if len(activedisplays) == 0:
		raise Exception("Couldn't find any monitors - did xrandr return something unintelligible?")
	if len(activedisplays) == 1:
		print "Found exactly one monitor; using", activedisplays[0]
		return activedisplays[0]
	chosen_monitor = prompt(strings, prompt_string="Please select a monitor:", default=len(activedisplays)-1)
	return activedisplays[chosen_monitor]

def choose_sound_device():
	"""Lets the user pick an ALSA device to use for recording."""
	# Actually it's easier to punt and use "default" or "pulse" and choose through pavucontrol
	pass


def choose_camera_device():
	"""Lets the user pick an external video device to use from /dev/video*"""
	videodevs = ["/dev/" + x for x in os.listdir("/dev/") if x.startswith("video") ]
	#print videodevs
	if len(videodevs) == 1:
		print "Found exactly one video device; using", videodevs[0]
		return videodevs[0]
	if len(videodevs) == 0:
		raise Exception("Found zero video devices - try plugging in a webcam and checking for /dev/video0?")

	# Do ridiculous ioctl stuff to extract the name of the device.
	# Here be dragons.
	import fcntl
	_IOC_NRBITS   =  8
	_IOC_TYPEBITS =  8
	_IOC_SIZEBITS = 14
	_IOC_DIRBITS  =  2

	_IOC_NRSHIFT = 0
	_IOC_TYPESHIFT =(_IOC_NRSHIFT+_IOC_NRBITS)
	_IOC_SIZESHIFT =(_IOC_TYPESHIFT+_IOC_TYPEBITS)
	_IOC_DIRSHIFT  =(_IOC_SIZESHIFT+_IOC_SIZEBITS)

	_IOC_NONE = 0
	_IOC_WRITE = 1
	_IOC_READ = 2
	def _IOC(direction,type,nr,size):
		return (((direction)  << _IOC_DIRSHIFT) |
			((type) << _IOC_TYPESHIFT) |
			((nr)   << _IOC_NRSHIFT) |
			((size) << _IOC_SIZESHIFT))
	def _IOR(type, number, size):
		return _IOC(_IOC_READ, type, number, size)
	def _IOW(type, number, size):
		return _IOC(_IOC_WRITE, type, number, size)

	sizeof_struct_v4l2_capability = (16 + 32 + 32 + 4 + 4 + 16)
	VIDIOC_QUERYCAP = _IOR(ord('V'),  0, sizeof_struct_v4l2_capability)

	import array
	import struct
	emptybuf = " " * (16 + 32 + 32 + 4 + 4 + 16) # sizeof(struct v4l2_capability)
	buf = array.array('c', emptybuf)
	cameranames = []
	for dev in videodevs:
		camera_dev = open(dev, "rw")
		camera_fd = camera_dev.fileno()
		fcntl.ioctl(camera_fd, VIDIOC_QUERYCAP, buf, 1)
		#print buf[16:48]
		cameranames.append(buf[16:48].tostring())
		camera_dev.close()
	
	chooseargs = ["%s - %s" % (videodevs[i], cameranames[i]) for i in xrange(len(videodevs))]
	index_chosen = prompt(chooseargs, prompt_string="Select a video camera to use for recording:")
	return videodevs[index_chosen]


if __name__ == "__main__":
	# 0) Sanity check and find binaries.
	# Check that all the executables that we want to run actually exist.
	gstbinary = None
	recordmydesktopbinary = None
	xrandrbinary = None
	for folder in os.environ["PATH"].split(":"):
		# Find gst-launch.
		if not gstbinary:
			for binaryname in [os.path.join(folder, b) for b in ["gst-launch", "gst-launch0.10"] ]:
				if os.path.exists(binaryname) and os.stat(binaryname).st_mode & 1:
					print "Found gst-launch at", binaryname
					gstbinary = binaryname
					break
		# Find recordmydesktop.
		if not recordmydesktopbinary:
			for binaryname in [os.path.join(folder, b) for b in ["recordmydesktop"] ]:
				if os.path.exists(binaryname) and os.stat(binaryname).st_mode & 1:
					print "Found recordmydesktop at", binaryname
					recordmydesktopbinary = binaryname
					break
		# Find xrandr.
		if not xrandrbinary:
			for binaryname in [os.path.join(folder, b) for b in ["xrandr"] ]:
				if os.path.exists(binaryname) and os.stat(binaryname).st_mode & 1:
					print "Found recordmydesktop at", binaryname
					xrandrbinary = binaryname
					break

	if not gstbinary:
		raise Exception("I couldn't find gst-launch.  Try installing gstreamer.")
	if not recordmydesktopbinary:
		raise Exception("I couldn't find recordmydesktop.  Try installing recordmydesktop.")
	if not xrandrbinary:
		raise Exception("I couldn't find xrandr.  Try installing xorg-x11 (how are you running this anyway? \(O_o)/ )")
	# 1) Pick a monitor to record.  Screenscrape xrandr to find what outputs are
	#    currently being driven, present multiple-choice menu.
	monitor_specs = choose_monitor()

	# 2) Pick a sound device to use.  Or we can punt and use "pulse", and have the user
	#    use pavucontrol to pick the appropriate underlying audio device.
	#sound_dev = choose_sound_device()
	sound_dev = "default"

	# 3) Pick an external video device to use from /dev/video*
	camera_dev = choose_camera_device()
	print "Using:"
	print "\tMonitor: %s - %sx%s @ (%s, %s)" % monitor_specs
	print "\tSound:   %s" % sound_dev
	print "\tCamera:  %s" % camera_dev

	# 4) Create a subfolder for this recording, named the current ISO8601 timestamp
	import datetime
	foldername = datetime.datetime.today().isoformat()
	print "Created folder", foldername
	os.mkdir(foldername)

	# 5) Try to inhibit screensaver.
	session_bus = None
	inhibit_cookie = None
	if dbus_enabled:
		try:
			print "Inhibiting screensaver..."
			session_bus = dbus.SessionBus()
			screensaver = session_bus.get_object("org.freedesktop.ScreenSaver", "/ScreenSaver")
			inhibit_cookie = screensaver.Inhibit("recorder.py", "Recording a screencast.")
		except:
			# If we can't find anything that implements the freedesktop.org
			# screensaver interface, don't bother trying to inhibit
			dbus_enabled = False

	# 6) Launch recordmydesktop for the appropriate monitor and sound device.
	args = "%s -o %s/screencast.ogv --width %s --height %s --device %s" % (recordmydesktopbinary, foldername, monitor_specs[1], monitor_specs[2], sound_dev)
	if int(monitor_specs[3]) > 0:
		args = args + " -x %s" % monitor_specs[3]
	if int(monitor_specs[4]) > 0:
		args = args + " -y %s" % monitor_specs[4]
	print args
	rmd_process = subprocess.Popen(args.split())

	# 7) Launch gstreamer recording the video from the webcam.
	# gst-launch v4l2src ! video/x-raw-yuv,width=640,height=480 ! ffmpegcolorspace ! theoraenc ! oggmux ! filesink location=testing.ogg
	# "gst-launch-0.10 
	args = [gstbinary]
	args.extend(("v4l2src device=%s ! video/x-raw-yuv,width=640,height=480 ! ffmpegcolorspace ! theoraenc ! oggmux ! filesink location=%s/camera.ogv" % (camera_dev, foldername)).split())
	gst_process = subprocess.Popen(args)

	print "recordmydesktop and gstreamer launched.  Press Control-C to stop recording."

	# 8) Wait for ^C.
	signal.signal(signal.SIGINT, signal_handler)
	signal.pause()

	# 9) If screensaver successfully inhibited before, uninhibit screensaver.
	if dbus_enabled and session_bus and inhibit_cookie:
		screensaver.UnInhibit(inhibit_cookie)

	# 10) Issue SIGINT to recordmydesktop and gstreamer.  Wait for them to terminate.
	#     recordmydesktop hasn't been encoding on the fly, so it could take a while.
	gst_process.send_signal(signal.SIGINT)
	rmd_process.send_signal(signal.SIGINT)
	gst_process.wait()
	rmd_process.wait()

	# TODO: postprocessing: reencode to other video formats?

	# 11) Return success.
	print "Recording done; files are in", foldername

